
# import karamba and other modules needed for KClock to work

import karamba, sys, os, datetime, time, math, re

from PIL import Image, ImageFilter

# Default style to use on a new installation
DEFAULT_STYLE_NAME = 'qs'

# list of clocks for each Widget instance
analogClocks = {}

# ids used for command responses
ids = {'style':0, 'resize':0}


# Base widget to handle the widget handle
class BaseWidget:
	def __init__(self, widget):
		self.widget = widget


# Config class to reand and write information to and from the karamba config file
class Config(BaseWidget):
	def __init__(self, widget):
		BaseWidget.__init__(self, widget)

	def readBool(self, name, defValue):
		if defValue:
			strValue = 'True'
		else:
			strValue = 'False'
		value = karamba.readConfigEntry(self.widget, name)
		if value == None:
			value = strValue
		if value == 'True':
			return True
		else:
			return False

	def readString(self, name, defValue):
		value = karamba.readConfigEntry(self.widget, name)
		if value == None:
			value = defValue
		return value
		
	def readTuple(self, name, defValue):
		value = karamba.readConfigEntry(self.widget, name)
		if value == None:
			value = defValue
		else:
			items = re.match('\(\s?(\d+)\s?,\s?(\d+)\s?\)', value)
			if items != None:
				value = (int(items.group(1)), int(items.group(2)))
		return value

	def writeBool(self, name, value):
		if value:
			value = 'True'
		else: 
			value = 'False'
		karamba.writeConfigEntry(self.widget, name, value)
	
	def writeString(self, name, value):
		karamba.writeConfigEntry(self.widget, name, value)

	def writeTuple(self, name, value):
		karamba.writeConfigEntry(self.widget, name, "(%d,%d)" % value)

# Class to handle the image cache
class ImageCache:
	def __init__(self, path):
		self.path = path
		if not os.path.exists(self.path):
			os.makedirs(self.path)
		self.load()
	
	def load(self):
		# load the cache file names into memory
		self.files = []
		self.files = os.listdir(self.path)
		
		
	def isImage(self, name):
		# is this name in the cache?
		index = -1
		try:
			index = self.files.index(name)
		except ValueError:
			pass
		return (index >= 0)
	
	def addImage(self, name):
		# add the image name to the cache
		self.files.append(name)

	def removeFiles(self):
		# remove all of the image files from cache and remove from memory
		for file in self.files:
			try:
				os.unlink(self.path + '/' + file)
			except:
				pass
		self.files = []


# Class to manage a stlye
class Style(BaseWidget):
	def __init__(self, widget):
		BaseWidget.__init__(self, widget)
		self.themePath = karamba.getThemePath(widget)
		# get the list of available styles
		self.names = os.listdir(self.themePath + "/styles")

	def load(self, name):
		# load a style 
		self.name = name
		self.path = self.themePath + "styles/" + self.name
		self.files = {'Clock': 'Clock.png',
						'Hour': 'Hour.png',
						'Minute': 'Minute.png',
						'Second': 'Second.png'
					}
		# set up the image cache
		cachePath = os.path.expanduser('~/.superkaramba/KClock_cache/') + self.name
		self.imageCache = ImageCache(cachePath)
		
	def getImageFile(self, imageName):
		# get the full path and filename for this style
		return self.path + '/' + self.files[imageName]

		
	def nextName(self):
		name = self.name
		try:
			index = self.names.index(name)
			index = index + 1
			if index >= len(self.names):
				index = 0
			name = self.names[index]
		except:
			pass
		return name

#  Base class to handle the widget and an image
class BaseImage(BaseWidget):
	def __init__(self, widget, imageName):
		BaseWidget.__init__(self, widget)
		self.name = imageName
		self.image = None
		

	def loadStyle(self, style):
		# load the image based on the givin style
		self.style = style
		self.filename = style.getImageFile(self.name)
		if self.image == None:
			self.image = karamba.createImage(self.widget, 0, 0, self.filename)
		else:
			karamba.setImagePath(self.widget, self.image, self.filename)	
		self.size = karamba.getImageSize(self.widget, self.image)
		self.deg = -1
		
	def resize(self, size):
		# resize the image to the given size
		if self.size != size:
			if self.image != None:
				karamba.resizeImage(self.widget, self.image, size[0], size[1])
		self.size = size

	def __del__(self):
		# on deletion of this object remove the image from karamba
		if self.image != None:
			karamba.deleteImage(self.widget, self.image)

# Basic clock hand image class
class ClockHandImage(BaseImage):
	def __init__(self, widget, parent, imageName):
		BaseImage.__init__(self, widget, imageName)
		self.parent = parent
		self.isVisible = True
		
	def dropShadow( self, image, offset=(5,5), background=0xffffff, shadow=0xffffff, 
					border=8, iterations=3):
		"""
		Add a gaussian blur drop shadow to an image.  
		
		image       - The image to overlay on top of the shadow.
		offset      - Offset of the shadow from the image as an (x,y) tuple.  Can be
						positive or negative.
		background  - Background colour behind the image.
		shadow      - Shadow colour (darkness).
		border      - Width of the border around the image.  This must be wide
						enough to account for the blurring of the shadow.
		iterations  - Number of times to apply the filter.  More iterations 
						produce a more blurred shadow, but increase processing time.
		"""
		
		# Create the backdrop image -- a box in the background colour with a 
		# shadow on it.
		totalWidth = image.size[0] + abs(offset[0]) + 2*border
		totalHeight = image.size[1] + abs(offset[1]) + 2*border
		back = Image.new(image.mode, (totalWidth, totalHeight), background)
		
		# Place the shadow, taking into account the offset from the image
		shadowLeft = border + max(offset[0], 0)
		shadowTop = border + max(offset[1], 0)
		back.paste(shadow, [shadowLeft, shadowTop, shadowLeft + image.size[0], 
			shadowTop + image.size[1]] )
		
		# Apply the filter to blur the edges of the shadow.  Since a small kernel
		# is used, the filter must be applied repeatedly to get a decent blur.
		n = 0
		while n < iterations:
			back = back.filter(ImageFilter.BLUR)
			n += 1
			
		# Paste the input image onto the shadow backdrop  
		imageLeft = border - min(offset[0], 0)
		imageTop = border - min(offset[1], 0)
		back.paste(image, (imageLeft, imageTop))
		
		return back
	
	
	def rotateImage(self, deg ):
		# rotate the hand image 
		if self.deg != deg and self.isVisible:
			#get the rotated cache name of the image
			rotatedImageName = self.name + '_' + ( '%d' % deg) + '.png'
			if not self.style.imageCache.isImage(rotatedImageName):
				# if not in the cache then we need to create it using PIL
				img = Image.open(self.filename)
				# check to see if we need to resize the image before rotation
				if not (img.size == self.size):
					img = img.resize(self.size)
				img = img.rotate(360 - deg, Image.BICUBIC)
				
				#shadow = self.dropShadow(img, (10, 10))
				img.save(self.style.imageCache.path + '/' + rotatedImageName)
				self.style.imageCache.addImage(rotatedImageName)
			# set the new image as the one to use for this object
			karamba.setImagePath(self.widget, self.image, self.style.imageCache.path + '/' + rotatedImageName)
			# now move the image to the center
			self.centerImage()
			self.deg = deg
	
	def centerImage(self):
		# center the image in the middle of the parent BaseImage ( the Clock face)
		size = karamba.getImageSize(self.widget, self.image)
		x = (size[0] - self.parent.size[0]) / 2
		y = (size[1] - self.parent.size[1]) / 2
		karamba.moveImage(self.widget, self.image, -x, -y)
		
	def setIsVisible(self, value):
		# set this image to visible ?
		if self.image != None:
			if value:
				karamba.showImage(self.widget, self.image)
			else:
				karamba.hideImage(self.widget, self.image)
		self.isVisible = value
		
	def writeConfig(self):
		# write the visible setting to the config file
		config = Config(self.widget)
		config.writeBool(self.name + '_isVisible', self.isVisible)
		
	def readConfig(self):
		# read the visible setting from the config file
		config = Config(self.widget)
		self.setIsVisible(config.readBool(self.name + '_isVisible', True))

# The Second Hand Class
class SecondHand(ClockHandImage):
	def __init__(self, widget, parent):
		ClockHandImage.__init__(self,  widget, parent, 'Second')
		
	def move(self, time):
		ClockHandImage.rotateImage(self, time.second * 6)

# The Minute Hand Class
class MinuteHand(ClockHandImage):
	def __init__(self, widget, parent):
		ClockHandImage.__init__(self, widget, parent, 'Minute')
	
	def move(self, time):
		deg = (time.minute * 6)
		ClockHandImage.rotateImage(self, deg)

# The Hour Hand Class
class HourHand(ClockHandImage):
	def __init__(self, widget, parent):
		ClockHandImage.__init__(self, widget, parent, 'Hour')
	
	def move(self, time):
		deg = (time.hour * 30) + (time.minute * 0.5)
		ClockHandImage.rotateImage(self, deg)

# The Clock
class AnalogClock(BaseImage):
	def __init__(self, widget):
		BaseImage.__init__(self, widget, 'Clock')
		# load up the hands
		self.hourHand = HourHand(widget, self)
		self.minuteHand = MinuteHand(widget, self)
		self.secondHand = SecondHand(widget, self)
		# create a stlye
		self.style = Style(widget)		
		
		# load up the current style
		config = Config(self.widget)
		self.style.load(config.readString("styleName", DEFAULT_STYLE_NAME))
		
		# set the style to the clock face
		self.loadStyle(self.style)

		# set the style to the hands
		self.hourHand.loadStyle(self.style)
		self.minuteHand.loadStyle(self.style)
		self.secondHand.loadStyle(self.style)
		self.secondHand.readConfig()
		
		# find out the set size of this widget
		size = config.readTuple('size', self.size)
		if  size != self.size:
			# if the save size is different to the loaded size then resize
			self.resize(size)
		

	def tick(self):
		# called on each second to move the hands
		time = datetime.datetime.now().time()
		self.hourHand.move(time)
		self.minuteHand.move(time)
		self.secondHand.move(time)

	def writeConfig(self):
		# write the config info to the config file
		config = Config(self.widget)
		config.writeString('styleName', self.style.name)
		self.secondHand.writeConfig()
		config.writeTuple('size', self.size)

	def setStyleName(self, name):
		# change the style to the 'name'
		
		size = self.size
		# load in the stlye
		self.style.load(name)
		# set the style to the clock face
		self.loadStyle(self.style)
		# set the style to the hands
		self.hourHand.loadStyle(self.style)
		self.minuteHand.loadStyle(self.style)
		self.secondHand.loadStyle(self.style)
		if  size != self.size:
			# if the save size is different to the loaded size then resize
			self.resize(size)

	def resize(self, size):
		# called to resize the clock and the hands
		# resize the widget to show the new images
		karamba.resizeWidget(self.widget, size[0], size[1])	
		# resize the base image ( Clock face)
		BaseImage.resize(self, size)
		# reszie the hands
		self.hourHand.resize(size)
		self.minuteHand.resize(size)
		self.secondHand.resize(size)
		# clear the cache as the current cache has images in the incorrect size
		self.style.imageCache.removeFiles()


#this is called when your widget is initialized
def initWidget(widget):
	global analogClocks
	
	# creaate a new clock
	analogClock = AnalogClock(widget)
	# assign for this instance
	analogClocks[widget] = analogClock
	propertiesMeter = karamba.createImage(widget,67,70,'C.png')
	# resize the widget for this size clock
	karamba.resizeWidget(widget, analogClock.size[0], analogClock.size[1])
	
	# do a tick to display everything
	analogClock.tick()

	# set up the menus
	karamba.addMenuConfigOption(widget, "style", "Change Style...")
	karamba.addMenuConfigOption(widget, "styleNext", "Next Style")
	karamba.addMenuConfigOption(widget, "secondIsVisible", "Show Second Hand")
	karamba.addMenuConfigOption(widget, "resize", "Resize...")
	karamba.setMenuConfigOption(widget, "style", 0)
	karamba.setMenuConfigOption(widget, "styleNext", 0)
	karamba.setMenuConfigOption(widget, "resize", 0)
	if analogClock.secondHand.isVisible:
		karamba.setMenuConfigOption(widget, "secondIsVisible", 0)
	else:
		karamba.setMenuConfigOption(widget, "secondIsVisible", 0)

#This is called when your widget is closed.  You can use this to clean
#up open files, etc.  You don't need to delete text and images in your
#theme.  That is done automatically.  This callback is just for cleaning up
#external things.  Most people don't need to put anything here.
def widgetClosed(widget):
	global analogClocks
	
	# clear out this instance of the clock
	analogClocks[widget] = None

#this is called everytime your widget is updated
#the update inverval is specified in the .theme file
def widgetUpdated(widget):
	global analogClocks
	
	# update the hands
	analogClocks[widget].tick()
	
#This gets called everytime our widget is clicked.
#Notes:
#  widget = reference to our widget
#  x = x position (relative to our widget)
#  y = y position (relative to our widget)
#  botton = button clicked:
#                    1 = Left Mouse Button
#                    2 = Middle Mouse Button
#                    3 = Right Mouse Button, but this will never happen
#                        because the right mouse button brings up the
#                        Karamba menu.
#                    4,5 = Scroll wheel up and down
def widgetClicked(widget, x, y, button):
    pass

#This gets called everytime our widget is clicked.
#Notes
#  widget = reference to our widget
#  x = x position (relative to our widget)
#  y = y position (relative to our widget)
#  botton = button being held:
#                    0 = No Mouse Button
#                    1 = Left Mouse Button
#                    2 = Middle Mouse Button
#                    3 = Right Mouse Button, but this will never happen
#                        because the right mouse button brings up the
#                        Karamba menu.
def widgetMouseMoved(widget, x, y, button):
    #Warning:  Don't do anything too intensive here
    #You don't want to run some complex piece of code everytime the mouse moves
    pass


#This gets called when an item is clicked in a popup menu you have created.
#  menu = a reference to the menu
#  id = the number of the item that was clicked.
def menuItemClicked(widget, menu, id):
    pass

#This gets called when an item is clicked in the theme CONFIGURATION menu,
#not the popup menus that you create.
#  key = the reference to the configuration key that was changed
#  value = the new value (true or false) that was selected
def menuOptionChanged(widget, key, value):
	global analogClocks, ids
	
	# commands to send for the dialogs
	styleDialog = ["kdialog", "--title", "Styles", "--radiolist", "Please select a style:"]
	resizeDialog = ["kdialog", "--title", "Resize Clock", "--inputbox", "Please enter desired size (default 135):"]

	# get the current clock for this instance
	analogClock = analogClocks[widget]
	
	if key == "style":
		# get the list of styles and select the one we are using
		for styleName in analogClock.style.names:
			if styleName == analogClock.style.name:
				switch = 'on'
			else:
				switch = 'off'
			styleDialog += [styleName, styleName, switch]
		ids['style'] = karamba.executeInteractive(widget, styleDialog)
		karamba.setMenuConfigOption(widget, "style", 0)
	elif key == 'styleNext':
		# set the next style in the list
		analogClock.setStyleName(analogClocks[widget].style.nextName())
		analogClock.writeConfig()
		analogClock.tick()
		karamba.redrawWidget(widget)
		karamba.setMenuConfigOption(widget, "styleNext", 0)
	elif key == "secondIsVisible":
		# toggle the second hand is visible
		analogClock.secondHand.setIsVisible(not analogClock.secondHand.isVisible)
		analogClock.writeConfig()
		analogClock.tick()
	elif key == "resize":
		# request a resize input box
		ids['resize'] = karamba.executeInteractive(widget, resizeDialog)
		karamba.setMenuConfigOption(widget, "resize", 0)


#This gets called when a meter (image, text, etc) is clicked.
# NOTE you must use attachClickArea() to make a meter
# clickable.  
#  widget = reference to your theme
#  meter = the meter clicked
#  button = the button clicked (see widgetClicked for button numbers)
def meterClicked(widget, meter, button):
    pass

#This gets called when a command you have executed with executeInteractive() outputs something
#to stdout.  This way you can get the output of for example kdialog without freezing up the widget
#waiting for kdialog to end.
#  widget = reference to your theme
#  pid = process number of the program outputting (use this if you execute more than out process)
#  output = the text the program outputted to stdout
def commandOutput(widget, pid, output):
	global analogClocks, ids

	# called when a dialog box returns
	# get the correct clock for this instance
	analogClock = analogClocks[widget]
	if pid == ids['style']:
		# get the style name and set it
		name = output.replace('\n','')
		analogClock.setStyleName(name)
		analogClock.writeConfig()
		analogClock.tick()
		karamba.redrawWidget(widget)
	elif pid == ids['resize']:
		# get the entered in size
		# validate the size if out of bounds then reset to the default size
		value = int(output)
		if ( value < 10 or value > 1000):
			value = 135
		analogClock.resize(widget, (value, value))
		analogClock.writeConfig()
		analogClock.tick()


#This gets called when another theme sends a message.
#  widget = reference to your theme
#  caller = "pretty" name or regular name (name of theme file with a number attached) of sending theme
#  output = the text sent.
def themeNotify(widget, caller, message):
    pass

#This gets called when an item is dropped on this widget.
# NOTE you have to call acceptDrops() before your widget will accept drops.
#  widget = reference to your theme
#  dropText = the text of the dropped item (probably a URL to it's location in KDE)
#  x = the x position on the theme relative to the upper left corner
#  y = the y position on the theme relative to the upper left corner
def itemDropped(widget, dropText, x, y):
    pass


#This gets called when a new program is LOADING in KDE.  When it is done
#loading, startupRemoved() is called, followed by taskAdded().
#  widget = reference to your widget
#  task = A refence to the task that is starting.  
def startupAdded(widget, startup):
    pass

#This gets called when a new program is done LOADING in KDE.
#  widget = reference to your widget
#  task = A refence to the task that just finished loading.  
def startupRemoved(widget, startup):
    pass

#This is called every time a new task (program) is started in KDE.
#  widget = reference to your widget
#  task = A refence to the new task.  Call getTaskInfo() with this reference
#         to get the name, etc of this new task.
def taskAdded(widget, task):
    pass

#This is called everytime a task (program) is closed in KDE.
#  widget = reference to your widget
#  task = A refence to the task.  
def taskRemoved(widget, task):
    pass

#This is called everytime a different task gains focus (IE, the user clicks
#on a different window).  
#  widget = reference to your widget
#  task = A refence to the task.  Call getTaskInfo() with this reference
#         to get the name, etc of this new task.
def activeTaskChanged(widget, task):
    pass

#This is called everytime the systray you created with createSystray() is updated
def systrayUpdated(widget):
    pass

#This is called everytime the current desktop changes
#  widget = reference to your widget
#  desktop = the current desktop
def desktopChanged(widget, desktop):
    pass

#This is called everytime the wallpaper changes on a desktop 
#  widget = reference to your widget
#  desktop = the desktop whose wallpaper changed
def wallpaperChanged(widget, desktop):
    pass

#This is called everytime there is a key press in any focused input field
#  widget = reference to your widget
#  meter = reference to an input box
#  char = the key that was pressed
def keyPressed(widget, meter, char):
    pass

# This will be printed when the widget loads.
print "KClock Loaded!"
