#!/usr/local/bin/python
#
# $Id: //projects/qstat/qstat.py#1 $ $Date: 2003/02/24 $

"""
Python wrapper for QStat, a realtime game server status application.
"""

__program__ = 'quakestat'
__version__ = '1.0'
__url__ = 'http://www.alcyone.com/pyos/qstat/'
__author__ = 'Erik Max Francis <max@alcyone.com>'
__copyright__ = 'Copyright (C) 2003 Erik Max Francis'
__license__ = 'GPL'


import getopt
import os
import sys
import xml.dom
import xml.dom.minidom

from xml.dom import Node


DEFAULT_PATH_NAME = 'quakestat'
ENCODING = 'iso-8859-1'


def getAttribute(node, name):
    attribute = node.getAttribute(name)
    return attribute.encode(ENCODING, 'ignore')

def getText(parentNode, tagName):
    result = ''
    node = None
    for childNode in parentNode.childNodes:
        if childNode.nodeType == Node.ELEMENT_NODE and \
           childNode.tagName == tagName:
            node = childNode
            break
    else:
        return None
    for textNode in node.childNodes:
        assert textNode.nodeType == Node.TEXT_NODE
        result += textNode.nodeValue
    return result.encode(ENCODING, 'ignore')

def getNumericText(parentNode, tagName):
    text = getText(parentNode, tagName)
    if text is not None:
        return int(text)
    else:
        return None


class Error(Exception): pass


class Player:

    """An abstraction of a player, contained within the Server.players
    list."""

    def __init__(self):
        self.name = None
        self.score = None
        self.ping = None
        self.team = None
        self.skin = None
        self.mesh = None
        self.face = None


class Server:

    """An abstraction of a server (returned from the
    Application.populate method."""

    def __init__(self):
        # Attributes
        self.type = None
        self.address = None
        self.status = None
        # Subnodes
        self.hostName = None
        self.name = None
        self.gameType = None
        self.map = None
        self.numPlayers = None
        self.maxPlayers = None
        self.ping = None
        self.retries = None
        self.players = []


class Application:

    """An abstraction of a qstat application."""

    ServerFactory = Server
    PlayerFactory = Player
    
    def getAttribute(node, name):
        attribute = node.getAttribute(name)
        return attribute.encode(ENCODING, 'ignore')
    getAttribute = staticmethod(getAttribute)
    
    def getText(parentNode, tagName):
        result = ''
        node = None
        for childNode in parentNode.childNodes:
            if childNode.nodeType == Node.ELEMENT_NODE and \
               childNode.tagName == tagName:
                node = childNode
                break
        else:
            return None
        for textNode in node.childNodes:
            assert textNode.nodeType == Node.TEXT_NODE
            result += textNode.nodeValue
        return result.encode(ENCODING, 'ignore')
    getText = staticmethod(getText)

    def getNumericText(parentNode, tagName):
        text = getText(parentNode, tagName)
        if text is not None:
            return int(text)
        else:
            return None
    getNumericText = staticmethod(getNumericText)
    
    def __init__(self, path=DEFAULT_PATH_NAME):
        self.path = path

    def verify(self):
        """Verify that the application is accessible."""
        try:
            pipe = os.popen("%s 2> /dev/null" % self.path)
            try:
                if not pipe.read():
                    raise Error, "QStat application not accessible"
            finally:
                pipe.close()
        except IOError:
            raise Error, "QStat application communication interrupted"

    def run(self, type, name, port=None):
        """Run the application."""
        if port is None:
            address = name
        else:
            address = '%s:%d' % (name, port)
        pipe = os.popen("%s -P -xml -%s %s 2> /dev/null" % \
                        (self.path, type, address))
        try:
            return self.populate(pipe)
        finally:
            pipe.close()

    def populate(self, file):
        """Return a server object, given a file-like stream representing
        QStat ."""
        server = self.ServerFactory()
        dom = xml.dom.minidom.parse(file)
        try:
            rootNode = dom.childNodes[0]
            assert rootNode.nodeType == Node.ELEMENT_NODE
            assert rootNode.nodeName == 'qstat'
            serverNodes = dom.getElementsByTagName('server')
            assert len(serverNodes) == 1
            serverNode = serverNodes[0]
            server.type = self.getAttribute(serverNode, 'type')
            server.address = self.getAttribute(serverNode, 'address')
            server.status = self.getAttribute(serverNode, 'status') == 'UP'
            server.hostName = self.getText(serverNode, 'hostname')
            server.name = self.getText(serverNode, 'name')
            server.gameType = self.getText(serverNode, 'gametype')
            server.map = self.getText(serverNode, 'map')
            server.numPlayers = self.getNumericText(serverNode, 'numplayers')
            server.maxPlayers = self.getNumericText(serverNode, 'maxplayers')
            server.ping = self.getNumericText(serverNode, 'ping')
            server.retries = self.getNumericText(serverNode, 'retries')
            playersNodes = serverNode.getElementsByTagName('players')
            if playersNodes:
                assert len(playersNodes) == 1
                playersNode = playersNodes[0]
                playerNodes = playersNode.getElementsByTagName('player')
                for playerNode in playerNodes:
                    player = self.PlayerFactory()
                    player.name = self.getText(playerNode, 'name')
                    player.score = self.getNumericText(playerNode, 'score')
                    player.ping = self.getNumericText(playerNode, 'ping')
                    player.team = self.getText(playerNode, 'team')
                    player.skin = self.getText(playerNode, 'skin')
                    player.mesh = self.getText(playerNode, 'mesh')
                    player.face = self.getText(playerNode, 'face')
                    server.players.append(player)
            return server
        finally:
            dom.unlink()

def usage():
    print >> sys.stderr, """\
Usage: %s [options] <server type> <server address>...

Options:
  -V --version      Print version and exit
  -h --help         Print help and exit
  -q --qstat=path   Specify a non-standard path to the QStat application
  -p --port=port    Specify a non-standard port for the server type

The server type is the code used in QStat; run qstat for more information.
""" % sys.argv[0],

def main():
    path = DEFAULT_PATH_NAME
    port = None
    # Parse the options.
    pairs, remainder = getopt.getopt(sys.argv[1:], 'Vhq:p:', ['version', 'help', 'qstat=', 'port='])
    for option, argument in pairs:
        if option in ('-V', '--version'):
            sys.stderr.write("%s version %s\n" % (__program__, __version__))
            return
        elif option in ('-h', '--help'):
            usage()
            return
        elif option in ('-q', '--qstat'):
            path = argument
        elif option in ('-p', '--port'):
            port = int(argument)
    if not remainder:
        usage()
        return
    if len(remainder) < 2:
        raise ValueError, "need at least two arguments"
    serverType, serverNames = remainder[0], remainder[1:]
    # Do it.
    application = Application(path)
    application.verify()
    ok = 1
    for serverName in serverNames:
        server = application.run(serverType, serverName, port)
        print "Server: %s (%s, type %s)" % \
              (('DOWN', 'UP')[server.status], server.address, server.type)
        if server.status:
            print "Host name: %s" % server.hostName
            print "Name: %s" % server.name
            print "Game type: %s" % server.gameType
            print "Map: %s" % server.map
            print "Players: %d/%d" % (server.numPlayers, server.maxPlayers)
            print "Ping (retries): %d (%d)" % (server.ping, server.retries)
            if server.numPlayers:
                print "Players:"
                for player in server.players:
                    print "\t%s (%d)" % (player.name, player.score)
        if len(serverNames) > 1:
            print
        if not server.status:
            ok = 0
    sys.exit(not ok)

if __name__ == '__main__': main()
